/*
esp8266 controlled relay board
-> 6 relays NO/NC
-> 2 assignable buttons
-> selectable and assignable groups
-> web interface (switching, assignment)
-> websocket control (switching, assignment)
-> UART control (switching)
-> HTTP GET control (switching)

-> schematics: https://www.analogmensch.de/leben_meinung/20230202/RelayBoard(schematic).pdf
*/

//libraries
#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <ESPAsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <EEPROM.h>
#include "WirelessData.h"
#include "WebInterface.h"

//inputs and outputs
#define button1 16
#define button2 A0
#define relay1 15
#define relay2 13
#define relay3 5
#define relay4 4
#define relay5 14
#define relay6 12
#define LED 2

//settings
int groupAmount = 25;  //amount of group IDs, min 1, max 99 (also check group state array)
int pulseDelay = 125;  //delay time for pulse mode

//button value array
int buttonValue[2] = {0, 0};

//button state array
int buttonState[2] = {0, 0};

//button group array
int buttonGroup[2] = {0, 0};

//relay value array
int relayValue[6] = {relay1, relay2, relay3, relay4, relay5, relay6};

//relay states array
int relayState[6] = {0, 0, 0, 0, 0, 0};

//relay groups array
byte groupState[26] = {
  0b00000000,  //group 0
  0b00000000,  //group 1
  0b00000000,  //group 2
  0b00000000,  //group 3
  0b00000000,  //group 4
  0b00000000,  //group 5
  0b00000000,  //group 6
  0b00000000,  //group 7
  0b00000000,  //group 8
  0b00000000,  //group 9
  0b00000000,  //group 10
  0b00000000,  //group 11
  0b00000000,  //group 12
  0b00000000,  //group 13
  0b00000000,  //group 14
  0b00000000,  //group 15
  0b00000000,  //group 16
  0b00000000,  //group 17
  0b00000000,  //group 18
  0b00000000,  //group 19
  0b00000000,  //group 20
  0b00000000,  //group 21
  0b00000000,  //group 22
  0b00000000,  //group 23
  0b00000000,  //group 24
  0b00000000   //group 25
};

//HTTP GET parameter names
const char* httpParameterOff = "off";
const char* httpParameterOn = "on";
const char* httpParameterGroup = "number";

//HTTP GET parameter strings
String httpMessageOff;
String httpMessageOn;
String httpMessageGroup;

//web server object on port 80
AsyncWebServer server(80);
//web socket object on sub page /websocket
AsyncWebSocket ws("/websocket");

void websocketChange(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type, void *arg, uint8_t *data, size_t len) {
  switch (type) {
    case WS_EVT_CONNECT: { 
        Serial.printf("websocketConnect:client#%u\nclient#%uAdress:%s\n",client->id(),client->id(),client->remoteIP().toString().c_str());
        websocketSet(client->id(),"clientID:#" + String(client->id()));
        websocketSet(client->id(),"buttonState:" + String(buttonState[0]) + String(buttonState[1]));
        String sRelayValue;
        for (int i=0;i<=5;i++) {
          sRelayValue += String(relayState[i]);
        }
        websocketSet(client->id(),"relayState:" + sRelayValue);
        websocketSet(client->id(),"buttonGroup:1:" + String(buttonGroup[0]) + "/2:" + String(buttonGroup[1]));
        String sState;
        for (int i=0;i<=groupAmount;i++) {
          String sStatePart;
          for (int j=0;j<=5;j++) {
            sStatePart += bitRead(groupState[i], j);
          }
          sState += String(i) + ":" + sStatePart;
          if (i < groupAmount) {
            sState += "/";
          }
        }
        websocketSet(client->id(),"relayGroup:" + sState);
      }
      break;
    case WS_EVT_DISCONNECT: {
        Serial.printf("websocketDisconnect:#%u\n",client->id());
      }
      break;
    case WS_EVT_DATA: {
        AwsFrameInfo *info = (AwsFrameInfo*)arg;
        if (info->final && info->index == 0 && info->len == len && info->opcode == WS_TEXT) {
          data[len] = 0;
          String sData = String((char*)data);
          Serial.print("websocketData#" + String(client->id()) + ":" + sData + "\n");  
          websocketSet(client->id(), "websocketData#" + String(client->id()) + ":" + sData);
          websocketGet(client->id(), sData);
        }
      }
      break;
    case WS_EVT_PONG: {
        Serial.print("websocketPong\n");
      } 
      break;
    case WS_EVT_ERROR: {
        Serial.print("websocketError\n");
      }
      break;
  }
}

void websocketGet(int id, String message) {  //incomming messages from websocket clients
  bool messageError = false;
  if (message.startsWith("relay:")) {  //message for relay switching
    bool toggle = false;
    bool pulse = false;
    int iNumberRelay;
    int iNumberState;
    char cNumberRelay = message[6];
    char cNumberState = message[8];
    if (message.length() == 9 && (cNumberRelay >= '1' && cNumberRelay <= '6') && (cNumberState == '0' || cNumberState == '1' || cNumberState == 't' || cNumberState == 'p') && message[7] == ':') {  //check message length, relay number, relay state and delimeter
      iNumberRelay = String(cNumberRelay).toInt();
      if (cNumberState == '0' || cNumberState == '1') {
        iNumberState = String(cNumberState).toInt();
      }
      else if (cNumberState == 't') {
        toggle = true;
      }
      else if (cNumberState == 'p') {
        pulse = true;
      }
    }
    else {
      messageError = true;
    }
    if (!messageError) { 
      if (!toggle && !pulse) {  //changeover mode
        relayState[iNumberRelay - 1] = iNumberState;
        relayChange();
      }
      else if (toggle && !pulse) {  //pulse mode
        int state = digitalRead(relayValue[iNumberRelay - 1]);
        relayState[iNumberRelay - 1] = !state;
        relayChange();
      }
      else if (!toggle && pulse) {  //pulse mode
        int state = digitalRead(relayValue[iNumberRelay - 1]);
        relayState[iNumberRelay - 1] = !state;
        relayChange();
        unsigned long time = millis();
        while(millis() - pulseDelay <= time) {
        }
        relayState[iNumberRelay - 1] = state;
        relayChange();
      }
    }
    else {
      Serial.print("websocketError\n");
    }
  }
  else if (message.startsWith("button:")) {  //message for group select
    int iNumberButton;
    int iNumberGroup;
    char cNumberButton = message[7];
    char cNumberGroupX = message[9];
    char cNumberGroupY = message[10];
    if (message.length() >= 10 && message.length() <= 11 && (cNumberButton == '1' || cNumberButton == '2') && (cNumberGroupX >= '0' && cNumberGroupX <= '9') && ((cNumberGroupY >= '0' && cNumberGroupY <= '9') || message.length() == 10) && message[8] == ':') {  //check message length, button numbers, group numbers and delimiter
      if (message.length() == 10) {  //if one digit group number
        iNumberButton = String(cNumberButton).toInt();
        iNumberGroup = String(cNumberGroupX).toInt();
      }
      else if (message.length() == 11) {  //two digit group number
        iNumberButton = String(cNumberButton).toInt();
        iNumberGroup = (String(cNumberGroupX) + String(cNumberGroupY)).toInt();
      }
    }
    else {
      messageError = true;
    }
    if (iNumberGroup > groupAmount) {  //check for maximum group amount
      messageError = true;
    }
    if (!messageError) {
      groupChangeButton(iNumberButton, iNumberGroup);
    }
    else {
      Serial.print("websocketError\n");
    }
  }
  else if (message.startsWith("group:get:")) {  //message for group select
    int iNumberGroup;
    char cNumberGroupX = message[10];
    char cNumberGroupY = message[11];
    if (message.length() >= 11 && message.length() <= 12 && (cNumberGroupX >= '0' && cNumberGroupX <= '9') && ((cNumberGroupY >= '0' && cNumberGroupY <= '9') || message.length() == 11)) {  //check message length and group numbers
      if (message.length() == 11) {  //if one digit group number
        iNumberGroup = String(cNumberGroupX).toInt();
      }
      else if (message.length() == 12) {  //two digit group number
        iNumberGroup = (String(cNumberGroupX) + String(cNumberGroupY)).toInt();
      }
    }
    else {
      messageError = true;
    }
    if (iNumberGroup > groupAmount) {  //check for maximum group amount
      messageError = true;
    }
    if (!messageError) {
      groupSelect(iNumberGroup);
    }
    else {
      Serial.print("websocketError\n");
    }
  }
  else if (message.startsWith("group:set:")) {  //message for group select
    int iNumberGroup;
    char cNumberGroupX = message[10];
    char cNumberGroupY = message[11];
    byte state = 0b00000000;
    if (message.length() >= 18 && message.length() <= 19 && (cNumberGroupX >= '0' && cNumberGroupX <= '9') && ((cNumberGroupY >= '0' && cNumberGroupY <= '9') || message.length() == 18)) {  //check message length and group numbers
      if (message.length() == 18 && message[11] == ':') {  //if one digit group number
        iNumberGroup = String(cNumberGroupX).toInt();
        for (int i=12;i<=17;i++) {
          if (message[i] == '0' || message[i] == '1') {
            bitWrite(state, i - 12, (String(message[i])).toInt());
          }
          else {
            messageError = true;
          }
        }
      }
      else if (message.length() == 19 && message[12] == ':') {  //two digit group number
        iNumberGroup = (String(cNumberGroupX) + String(cNumberGroupY)).toInt();
        for (int i=13;i<=18;i++) {
          if (message[i] == '0' || message[i] == '1') {
            bitWrite(state, i - 13, (String(message[i])).toInt());
          }
          else {
            messageError = true;
          }
        }
      }
    }
    else {
      messageError = true;
    }
    if (iNumberGroup > groupAmount) {  //check for maximum group amount
      messageError = true;
    }
    if (!messageError) {
      groupChangeRelay(iNumberGroup, state);
    }
    else {
      Serial.print("websocketError\n");
    }
  }
  else if (message == "state") {  //message for state request
    websocketSet(id,"buttonState:" + String(buttonState[0]) + String(buttonState[1]));
    String sRelayValue;
    for (int i=0;i<=5;i++) {
      sRelayValue += String(relayState[i]);
    }
    websocketSet(id,"relayState:" + sRelayValue);
  }
}

void websocketSet(int id, String message) {  //outgoing messages to websocket client
  if (id == 0) {  //if id is 0, send to all clients
    ws.textAll(message);
  }
  else {  //if is is not 0, send to specified client
    ws.text(id, message);
  }
}

void setup() {
  //UART configuration
  Serial.begin(9600);
  Serial.setTimeout(50);

  //pin mode for buttons and LED
  pinMode(button1, INPUT);
  pinMode(button2, INPUT);
  pinMode(LED, OUTPUT);

  //pin mode for relays and setting relay states to LOW
  for (int i = 0; i <= 5; i++) {
    pinMode(relayValue[i], OUTPUT);
    digitalWrite(relayValue[i], LOW);
  }

  //connect to WiFi
  WiFiConnect();

  //start server
  server.begin();

  //start web socket
  ws.onEvent(websocketChange);
  server.addHandler(&ws);

  //load EEPROM
  EEPROM.begin(groupAmount + 2);

  //read EEPROM
  for (int i=0;i<=1;i++) {  //read button assignment
    buttonGroup[i] = EEPROM.read(groupAmount + 1 + i);
  }
  for (int i=1;i<=groupAmount;i++) {  //read group assignment
    groupState[i] = EEPROM.read(i);
  }

  //HTTP request for /
  server.on("/", HTTP_GET, [](AsyncWebServerRequest*request) {
    request->redirect("/switching");
  });

  //HTTP request for /switching
  server.on("/switching", HTTP_GET, [](AsyncWebServerRequest*request) {
    request->send_P(200, "text/html", switching_html);
  });

  //HTTP request for /settings
  server.on("/settings", HTTP_GET, [](AsyncWebServerRequest* request) {
    //send group bytes and button assignment
    request->send_P(200, "text/html", settings_html);
  });

  //HTTP request for /change
  server.on("/change", HTTP_GET, [](AsyncWebServerRequest*request) {
    httpMessageOff = "";
    httpMessageOn = "";
    bool messageError = false;
    if (request->hasParam(httpParameterOff)) {  //check for off parameter
      Serial.print("httpChange\n");
      httpMessageOff = request->getParam(httpParameterOff)->value();
      if (httpMessageOff.length() >= 1 && httpMessageOff.length() <= 6) {
        for (int i=0;i<httpMessageOff.length() - 1;i++) {
          if (httpMessageOff[i] < '1' || httpMessageOff[i] > '6') {
            messageError = true;
          }
        }
      }
      else {
        messageError = true;
      }
    }
    if (request->hasParam(httpParameterOn)) {  //check for on parameter
      Serial.print("httpChange\n");
      httpMessageOn = request->getParam(httpParameterOn)->value();
      if (httpMessageOn.length() >= 1 && httpMessageOn.length() <= 6) {
        for (int i=0;i<httpMessageOn.length() - 1;i++) {
          if (httpMessageOn[i] < '1' || httpMessageOn[i] > '6') {
            messageError = true;
          }
        }
      }
      else {
        messageError = true;
      }
    }
    //check numbers strings for dupkicate numbers 
    int httpMessageOffLength = httpMessageOff.length();
    int httpMessageOnLength = httpMessageOn.length();
    if (httpMessageOffLength > 0 && httpMessageOnLength > 0) {
      for (int i=0;i<httpMessageOffLength;i++) {
        for (int j=0;j<httpMessageOnLength;j++) {
          if (httpMessageOff[i] == httpMessageOn[j]) {
            messageError = true;
          }
        }
      }
    }
    //write numbers to array, if messageError is false
    if (!messageError) {
      for (int i=0;i<httpMessageOffLength;i++) {
        int position = String(httpMessageOff[i]).toInt() - 1;
        relayState[position] = 0;
      }
      for (int i=0;i<httpMessageOnLength;i++) {
        int position = String(httpMessageOn[i]).toInt() - 1;
        relayState[position] = 1;
      }
      relayChange();
    }
    else {
      Serial.print("httpError\n");
    }
    request->redirect("/switching");
  });

  //HTTP request for /group
  server.on("/group", HTTP_GET, [](AsyncWebServerRequest*request) {
    bool messageError = false;
    httpMessageGroup = "";
    if (request->hasParam(httpParameterGroup)) {
      Serial.print("httpChange\n");
      int iNumberGroup;
      httpMessageGroup = request->getParam(httpParameterGroup)->value();
      if (httpMessageGroup.length() >= 1 && httpMessageGroup.length() <= 2) {
        for (int i=0;i<httpMessageGroup.length() - 1;i++) {
          if (httpMessageGroup[i] < '0' && httpMessageGroup[i] > '9') {
            messageError = true;
          }
        }
      }
      else {
        messageError = true;
      }
      iNumberGroup = httpMessageGroup.toInt();  
      if (iNumberGroup > groupAmount) {
        messageError = true;
      }
      if (!messageError) {
        groupSelect(iNumberGroup);
      }
      else {
        Serial.print("httpError\n");
      }
    }
    request->redirect("/switching");
  });

  //HTTP request for /state
  server.on("/state", HTTP_GET, [](AsyncWebServerRequest* request) {
    String sRelayValue;
    for (int i=0;i<=5;i++) {
      sRelayValue += String(relayState[i]);
    }
    String state = String(buttonState[0]) + String(buttonState[1]) + sRelayValue;
    char cState[state.length()+1];
    state.toCharArray(cState,state.length()+1);
    Serial.print("httpState:" + state + "\n");
    request->send(200, "text/plain", cState);
  });

  //HTTP 404 error
  server.onNotFound([](AsyncWebServerRequest *request){
    request->redirect("/");
  });
}

void loop() {
  //check WiFi connection
  if (WiFi.status() != WL_CONNECTED) {
    WiFiConnect();
  }

  //read UART
  if (Serial.available()) {
    String uart = Serial.readStringUntil('\n');
    if (uart.endsWith("\r")) {
      uart.remove(uart.length() - 1);
    }
    if (uart.startsWith("state")) {
      uartState(uart);
    }
    else if (uart.startsWith("change")) {
      uartChange(uart);
    }
  }

  //read buttons
  buttonChange();

  //interface loading time
  delay(25);
}

void uartState(String uart) {
  //send recieve validation
  Serial.print("uartState\n");
  //positions of parameter strings and numbers strings
  int positionButtonState = 0;
  int positionRelayState = 0;
  //output strings
  String sButtonState;
  String sRelayState;
  //check for button and relay command
  if (uart.indexOf("&button") >= 0) {
    positionButtonState = uart.indexOf("&button");
    sButtonState = "buttonState:" + String(buttonState[0]) + String(buttonState[1]) + "\n";
  }
  if (uart.indexOf("&relay") >= 0) {
    positionRelayState = uart.indexOf("&relay");
    String sRelayValue;
    for (int i=0;i<=5;i++) {
      sRelayValue += String(relayState[i]);
    }
    sRelayState = "relayState:" + sRelayValue + "\n";
  }
  //check if none, only button, only relay or both command are available
  if (positionButtonState == 0 && positionRelayState == 0) {  //none
    Serial.print("stateError\n");
  }
  else if (positionButtonState > 0 && positionRelayState == 0) {  //only button
    Serial.print(sButtonState);
  }
  else if (positionButtonState == 0 && positionRelayState > 0) {  //only relay
    Serial.print(sRelayState);
  }
  else if (positionButtonState > 0 && positionRelayState > 0) {  //both
    if (positionButtonState < positionRelayState) {  //button before relay
      Serial.print(sButtonState);
      Serial.print(sRelayState);
    }
    else if (positionButtonState > positionRelayState) {  //relay before button
      Serial.print(sRelayState);
      Serial.print(sButtonState);
    }
  }
}

void uartChange(String uart) {
  //send recieve validation
  Serial.print("uartChange\n");
  //positions of parameter strings and numbers strings
  int positionOffParameter = 0;
  int positionOnParameter = 0;
  int positionOffNumbers = 0;
  int positionOnNumbers = 0;
  //length of full uart string
  int uartLength = uart.length();
  //strings for saving numbers
  String numbersOff;
  String numbersOn;
  //change bool
  bool change = true;
  //numbers error bool
  bool numbersError = false;
  //check for off and on command
  if (uart.indexOf("&off=") >= 0) {
    positionOffParameter = uart.indexOf("&off=");
    positionOffNumbers = positionOffParameter + 5;
  }
  if (uart.indexOf("&on=") >= 0) {
    positionOnParameter = uart.indexOf("&on=");
    positionOnNumbers = positionOnParameter + 4;
  }
  //check if none, only off, only on or both command are available
  if (positionOffParameter == 0 && positionOnParameter == 0) {  //none
    change = false;
  }
  else if (positionOffParameter > 0 && positionOnParameter == 0) {  //only off
    numbersOff = uart.substring(positionOffNumbers,uartLength);
  }
  else if (positionOffParameter == 0 && positionOnParameter > 0) {  //only on
    numbersOn = uart.substring(positionOnNumbers,uartLength);
  }
  else if (positionOffParameter > 0 && positionOnParameter > 0) {  //off and on
    if (positionOffParameter < positionOnParameter) {  //off before on
      numbersOff = uart.substring(positionOffNumbers,positionOnParameter);
      numbersOn = uart.substring(positionOnNumbers,uartLength);
    }
    if (positionOffParameter > positionOnParameter) {  //on before off
      numbersOff = uart.substring(positionOffNumbers,uartLength);
      numbersOn = uart.substring(positionOnNumbers,positionOffParameter);
    }
  }
  //check numbers strings for invalid characters
  for (int i=0;i<numbersOff.length();i++) {
    char numbersCheck = numbersOff[i];
    if (numbersCheck < '1' || numbersCheck > '6') {
      numbersError = true;
    }
  }
  for (int i=0;i<numbersOn.length();i++) {
    char numbersCheck = numbersOn[i];
    if (numbersCheck < '1' || numbersCheck > '6') {
      numbersError = true;
    }
  }
  //check numbers strings for duplicate numbers 
  int numbersOffLength = numbersOff.length();
  int numbersOnLength = numbersOn.length();
  if (numbersOffLength > 0 && numbersOnLength > 0) {
    for (int i=0;i<numbersOffLength;i++) {
      for (int j=0;j<numbersOnLength;j++) {
        if (numbersOff[i] == numbersOn[j]) {
          numbersError = true;
        }
      }
    }
  }
  //write numbers to array, if numbersError is false
  if (change && !numbersError) {
    for (int i=0;i<numbersOffLength;i++) {
      int position = String(numbersOff[i]).toInt() - 1;
      relayState[position] = 0;
    }
    for (int i=0;i<numbersOnLength;i++) {
      int position = String(numbersOn[i]).toInt() - 1;
      relayState[position] = 1;
      }
    relayChange();
  }
  //...or send error message
  else {
    Serial.print("changeError\n");
  }
}

void buttonChange() {
  //read button values
  buttonValue[0] = digitalRead(button1);
  buttonValue[1] = analogRead(button2);
  //set button 1 state
  if (buttonValue[0] == 0 && !buttonState[0]) {
    buttonState[0] = true;
    Serial.print("buttonChange:" + String(buttonState[0]) + String(buttonState[1]) + "\n");
    websocketSet(0, "buttonState:" + String(buttonState[0]) + String(buttonState[1]));
    groupSelect(buttonGroup[0]);
  }
  else if (buttonValue[0] == 1 && buttonState[0]) {
    buttonState[0] = false;
    Serial.print("buttonChange:" + String(buttonState[0]) + String(buttonState[1]) + "\n");
    websocketSet(0, "buttonState:" + String(buttonState[0]) + String(buttonState[1]));
  }
  //set button 2 state
  if (buttonValue[1] < 511 && !buttonState[1]) {
    buttonState[1] = true;
    Serial.print("buttonChange:" + String(buttonState[0]) + String(buttonState[1]) + "\n");
    websocketSet(0, "buttonState:" + String(buttonState[0]) + String(buttonState[1]));
    groupSelect(buttonGroup[1]);
  }
  else if (buttonValue[1] > 511 && buttonState[1]) {
    buttonState[1] = false;
    Serial.print("buttonChange:" + String(buttonState[0]) + String(buttonState[1]) + "\n");
    websocketSet(0, "buttonState:" + String(buttonState[0]) + String(buttonState[1]));
  }
}

void relayChange() {
  String sChange;
  Serial.print("relayChange:");
  for (int i = 0; i <= 5; i++) {
    digitalWrite(relayValue[i], relayState[i]);
    sChange += relayState[i];
    Serial.print(String(relayState[i]));
  }
  Serial.print("\n");
  websocketSet(0, "relayState:" + sChange);
}

void groupChangeRelay(int group, byte state) {
  groupState[group] = state;
  if (EEPROM.read(group) != state) {
    EEPROM.write(group, state);
    EEPROM.commit();
  }
  String sState;
  for (int i=0;i<=5;i++) {
    sState += bitRead(state, i);
  }
  websocketSet(0, "groupChangeRelay:" + String(group) + ":" + sState);
}

void groupSelect(int group) {
  for (int i=0;i<=5;i++) {
    relayState[i] = bitRead(groupState[group], i);
  }
  websocketSet(0, "groupSelect:" + String(group));
  relayChange();
}

void groupChangeButton(int button, int group) {
  buttonGroup[button - 1] = group;
  if (EEPROM.read(groupAmount + button) != group) {
    EEPROM.write(groupAmount + button, group);
    EEPROM.commit();
  }
  websocketSet(0, "groupChangeButton:" + String(button) + ":" + String(group));
}

void WiFiConnect() {
  WiFi.begin(ssid, password);
  Serial.print("wifiConnect:");
  Serial.print(ssid);
  Serial.print("\n");
  while (WiFi.status() != WL_CONNECTED) {
    digitalWrite(LED, HIGH);
    delay(125);
    digitalWrite(LED, LOW);
    delay(125);
  }
  digitalWrite(LED, LOW);
  Serial.print("serverAddress:");
  Serial.print(WiFi.localIP());
  Serial.print("\n");
}